import datetime
import pandas as pd
import numpy as np
import plotly.express as px
import os


def create_treemap_data(df):
    basic_info_df = None
    if os.path.exists("company_basic_info.csv"):
        basic_info_df = pd.read_csv("company_basic_info.csv")
        basic_info_df["stock_id"] = basic_info_df["stock_id"].astype(str)
    else:
        url = "https://raw.githubusercontent.com/twfxjjbw/stockinfo/main/company_basic_info.csv"
        basic_info_df = pd.read_csv(url)
        basic_info_df["stock_id"] = basic_info_df["stock_id"].astype(str)
        basic_info_df.to_csv("company_basic_info.csv", index=False)
    basic_info_df["stock_id_name"] = basic_info_df["stock_id"] + basic_info_df["公司簡稱"]
    df = df.merge(
        basic_info_df[["stock_id", "stock_id_name", "產業類別", "市場別", "實收資本額(元)"]],
        how="left",
        on="stock_id",
    )
    df = df.rename(columns={"產業類別": "category", "市場別": "market", "實收資本額(元)": "base"})
    df = df.dropna(thresh=5)
    df["market_value"] = round(df["base"] / 10 * df["close"] / 100000000, 2)
    df["turnover_ratio"] = df["turnover"] / (df["turnover"].sum()) * 100
    df["country"] = "TW-Stock"
    return df


def plot_tw_stock_treemap(
    df, area_ind="turnover_ratio", item="return_ratio", color_scales="Temps"
):
    df = create_treemap_data(df)
    if df is None:
        return None
    df["custom_item_label"] = round(df[item], 2).astype(str)

    if area_ind not in ["market_value", "turnover", "turnover_ratio"]:
        return None

    if item in ["return_ratio"]:
        color_continuous_midpoint = 0
    else:
        color_continuous_midpoint = np.average(df[item], weights=df[area_ind])

    fig = px.treemap(
        df,
        path=["country", "market", "category", "stock_id_name"],
        values=area_ind,
        color=item,
        color_continuous_scale=color_scales,
        color_continuous_midpoint=color_continuous_midpoint,
        custom_data=["custom_item_label", "close", "turnover"],
        title=f"TW-Stock Market TreeMap"
        f"---area_ind:{area_ind}---item:{item} <span style='color:blue'>last_updated:{datetime.datetime.now().strftime('%Y/%m/%d %H:%M:%S')}</span><BR>",
        width=1600,
        height=800,
    )

    fig.update_traces(
        textposition="middle center",
        textfont_size=24,
        texttemplate="%{label}(%{customdata[1]})<br>%{customdata[0]}",
    )
    return fig



from datetime import timedelta
from retry import retry
from github import Github
from github import GithubException
from .aes_crypt import AESCrypt
from loguru import logger
import requests
import pandas as pd
import subprocess
import requests
import pickle
import os
import time
import json
import re


def get_stock_future_code() -> pd.DataFrame:
    """Fetch product codes from TFE website"""
    now = pd.Timestamp.now()
    year = now.year
    month = now.month
    url = "https://www.taifex.com.tw/cht/5/sSFFSP"
    payload = {
        "queryYear": str(year),
        f"queryMonth": str(month - 1).zfill(2),
        "button": "送出查詢",
    }
    df = pd.read_html(requests.post(url, data=payload).content)[1]
    標的證券代號 = list(set(df["標的證券代號"].tolist()))
    標的證券代號 = [str(i) for i in 標的證券代號 if len(str(i)) == 4]
    return 標的證券代號


def decrypt_github_file(repo, password, select_file):
    # Get file path
    file_path = get_file_path(repo, select_file)

    # Check if file exists
    if not file_path:
        exit()

    # Create the enc folder if it does not exist
    enc_folder = os.path.join(".", "enc")
    if not os.path.exists(enc_folder):
        os.makedirs(enc_folder)

    # Download and decrypt data from Github
    download_file(
        repo=repo,
        github_file_path=os.path.join(file_path),
        download_path=enc_folder,
    )

    # Decrypt data from downloaded file
    with open(os.path.join(enc_folder, select_file), "rb") as f:
        encrypted_data = (f.readline().strip(), f.readline().strip())

    # Decrypt data
    aes = AESCrypt(password)
    decrypted_data = aes.decrypt(encrypted_data)
    return decrypted_data.decode(encoding="utf-8")


def get_file_path(repo, select_file):
    """
    Retrieve the path of the input file from a GitHub repository.
    """
    contents = repo.get_contents("")
    while contents:
        file_content = contents.pop(0)
        if file_content.type == "dir":
            contents.extend(repo.get_contents(file_content.path))
        else:
            file_name = re.sub(r"^.+/", "", file_content.path)
            if file_name == select_file:
                return file_content.path

    print(f"File {select_file} not found in the repository.")
    return None


def get_file_list(repo):
    def get_file_names(contents):
        """
        A generator that recursively retrieves file names from a GitHub repository.
        """
        for file_content in contents:
            if file_content.type == "dir":
                contents.extend(repo.get_contents(file_content.path))
            else:
                file_name = re.sub(r"^.+\/", "", file_content.path)
                yield file_name

    all_files = list(get_file_names(repo.get_contents("")))
    return all_files


def get_github_repo(token, repo_name, timeout=60):
    """
    連接 Github API，獲取指定名稱和 Token 的 Github 倉庫對象
    """
    g = Github(token, timeout=timeout)
    user = g.get_user()
    repo = user.get_repo(repo_name)
    return repo


def get_dataset_df_data(dataset):
    url = f"https://github.com/twfxjjbw/stockinfo/raw/main/{dataset}.bin"
    response = requests.get(url)
    if response.status_code == 200:
        df = pickle.loads(response.content)
        return df
    else:
        raise Exception(f"Error downloading {dataset}: {response.status_code}")


def decrypt_data_from_github(
    filename, my_password, github_token, repo_name, branch=None
):
    repo = get_github_repo(token=github_token, repo_name=repo_name)

    # Create AESCrypt instance
    aes = AESCrypt(my_password)

    # Download encrypted file from Github
    download_file(repo=repo, github_file_path=filename, branch=branch)

    # Read encrypted data from file
    with open(filename, "rb") as f:
        encrypted_data = (f.readline().strip(), f.readline().strip())

    # Decrypt data
    decrypted_data = aes.decrypt(encrypted_data)

    # Deserialize decrypted data into Python object
    data_decrypt = json.loads(decrypted_data.decode("utf-8"))

    return data_decrypt


def branch_exists(repo, branch_name):
    try:
        repo.get_branch(branch_name)
        return True
    except GithubException:
        return False


def create_branch(repo, branch_name, from_branch):
    try:
        # Create the new branch from the specified branch (from_branch).
        repo.create_git_ref(
            ref=f"refs/heads/{branch_name}", sha=repo.get_branch(from_branch).commit.sha
        )
        return True
    except GithubException as e:
        print(e)
        return False


@retry(exceptions=Exception, tries=3, delay=10, backoff=2)
def download_file(repo, github_file_path, download_path=None, branch=None):
    filename = os.path.basename(github_file_path)
    if not download_path:
        download_path = os.getcwd()

    if branch:
        file_content = repo.get_contents(github_file_path, ref=branch)
    else:
        file_content = repo.get_contents(github_file_path)

    file_content_data = file_content.decoded_content
    mode = "wb" if isinstance(file_content_data, bytes) else "w"
    with open(os.path.join(download_path, filename), mode) as f:
        f.write(file_content_data)
    print(f"{filename} downloaded to {download_path}")


@retry(exceptions=Exception, tries=3, delay=10, backoff=2)
def update_file(repo, filename, folder_path="", branch=None):
    cwd = os.path.abspath(os.getcwd())
    try:
        if branch:
            contents = repo.get_contents(
                os.path.join(folder_path, os.path.basename(filename)), ref=branch
            )
        else:
            contents = repo.get_contents(
                os.path.join(folder_path, os.path.basename(filename))
            )
        sha = contents.sha
        logger.info(f"{filename} EXISTS")
    except Exception as e:
        sha = None
        logger.info(f"{filename} DOES NOT EXIST")

    content = None
    try:
        logger.info(f"Reading {filename}[+]")
        with open(filename, "rb") as f:
            file_content = f.read()

        if isinstance(file_content, str):
            with open(os.path.join(cwd, filename), "r") as f:
                content = f.read()
        elif isinstance(file_content, bytes):
            with open(os.path.join(cwd, filename), "rb") as f:
                content = f.read()
        logger.info(f"Reading {filename}[-]")
        logger.info(f"content length: {len(content)}, type: {type(content)}")
    except Exception as e:
        logger.error(f"Error reading {filename}: {e}")

    if sha:
        try:
            if branch:
                repo.update_file(
                    os.path.join(folder_path, os.path.basename(filename)),
                    "committing files",
                    content,
                    sha,
                    branch=branch,
                )
            else:
                repo.update_file(
                    os.path.join(folder_path, os.path.basename(filename)),
                    "committing files",
                    content,
                    sha,
                )
            logger.info(f"{filename} UPDATED")
        except Exception as e:
            logger.error(f"Error updating {filename}: {e}")
            raise e
    else:
        try:
            if branch:
                repo.create_file(
                    os.path.join(folder_path, os.path.basename(filename)),
                    "committing files",
                    content,
                    branch=branch,
                )
            else:
                repo.create_file(
                    os.path.join(folder_path, os.path.basename(filename)),
                    "committing files",
                    content,
                )
            logger.info(f"{filename} CREATED")
        except Exception as e:
            logger.error(f"Error creating {filename}: {e}")
            raise e


@retry(exceptions=Exception, tries=3, delay=2, backoff=2)
def delete_file(repo, filename, branch=None):
    """
    刪除指定的 Github 倉庫中的檔案
    """
    all_files = []
    if branch:
        contents = repo.get_contents("", ref=branch)
    else:
        contents = repo.get_contents("")

    while contents:
        file_content = contents.pop(0)
        if file_content.type == "dir":
            if branch:
                contents.extend(repo.get_contents(file_content.path, ref=branch))
            else:
                contents.extend(repo.get_contents(file_content.path))
        else:
            file = file_content
            all_files.append(file.path)

    # 要刪除的檔案路徑
    git_file = filename

    if git_file in all_files:
        # 如果已存在，刪除檔案
        if branch:
            file = repo.get_contents(git_file, ref=branch)
            repo.delete_file(file.path, "Delete files", file.sha, branch=branch)
        else:
            file = repo.get_contents(git_file)
            repo.delete_file(file.path, "Delete files", file.sha)
        print(f"{git_file} deleted")
    else:
        print(f"{git_file} does not exist.")


def get_execution_time(func):
    """
    裝飾器：計算函數執行時間
    """

    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        elapsed_time = timedelta(seconds=end_time - start_time)
        ms = int(elapsed_time.total_seconds() * 1000)
        print("Function {} took {} ms to execute.".format(func.__name__, ms))
        return result

    return wrapper


def get_fonts():
    # Run fc-list command and capture output
    output = subprocess.check_output(["fc-list"])

    # Decode output and split into lines
    output = output.decode("utf-8")
    lines = output.split("\n")

    # Extract font names from lines
    fonts = []
    for line in lines:
        if ":" in line:
            font = line.split(":")[0]
            fonts.append(font)

    # Extract font names from lines
    font_names = []
    for font in fonts:
        f = font.split("/")[-1]
        font_names.append(f)

    return font_names



import pandas as pd


class PriceMetrics:
    def __init__(self, prices):
        self.prices = prices
        self.base_price = prices[0]
        self.min_price = min(prices)
        self.max_price = max(prices[1:])
        self.max_price_before_min = self._get_max_price_before_min()
        self.mae = self._calculate_mae()
        self.bmfe = self._calculate_bmfe()
        self.gmfe = self._calculate_gmfe()
        self.edge_ratio = self._calculate_edge_ratio()

    def _get_max_price_before_min(self):
        if (
            self.min_price is None
            or self.prices[: self.prices.index(self.min_price)] == []
        ):
            return None
        return max(self.prices[: self.prices.index(self.min_price)])

    def _calculate_mae(self):
        if self.min_price is None or self.min_price >= self.base_price:
            return 0
        return (self.min_price - self.base_price) / self.base_price * 100

    def _calculate_bmfe(self):
        if (
            self.max_price_before_min is None
            or self.base_price >= self.max_price_before_min
        ):
            return 0
        return ((self.max_price_before_min - self.base_price) / self.base_price) * 100

    def _calculate_gmfe(self):
        if self.max_price is None or self.max_price <= self.base_price:
            return 0
        return ((self.max_price - self.base_price) / self.base_price) * 100

    def _calculate_edge_ratio(self):
        if self.mae == 0:
            return self._calculate_gmfe()
        return ((self.gmfe - abs(self.mae)) / abs(self.mae)) * 100

    def get_max_drawdown(self):
        df = pd.DataFrame(self.prices, columns=["Close"])
        df["Max"] = df["Close"].cummax()
        df["Drawdown"] = df["Close"] / df["Max"] - 1
        df["Max Drawdown"] = df["Drawdown"].cummin()
        mdd = df["Max Drawdown"].min()
        return mdd * 100


if __name__ == "__main__":
    # prices = [100, 90, 80, 70, 60, 50, 40, 30, 20, 10]
    prices = [10, 20, 15, 25, 18, 30, 12]
    metrics = PriceMetrics(prices)
    print(metrics.mae)
    print(metrics.bmfe)
    print(metrics.gmfe)
    print(metrics.edge_ratio)
    print(metrics.get_max_drawdown())



import datetime
import pandas as pd


class ContractFinder:
    def __init__(self, api):
        self.api = api

    def get_latest_option_contract(self):
        wednesday_time = self._get_next_wednesday()
        contracts_df = self._get_tx_options_contracts()
        return self._get_next_contract(contracts_df, wednesday_time)

    def _get_next_wednesday(self):
        today = datetime.datetime.today()
        wednesday = (today + datetime.timedelta(days=(2 - today.weekday()))).date()
        wednesday_time = datetime.datetime.combine(
            wednesday, datetime.datetime.min.time()
        ) + datetime.timedelta(
            hours=13, minutes=45, seconds=0
        )  # Adjusted to 13:45 due to 14:50 activation time
        return wednesday_time

    def _get_tx_options_contracts(self):
        # Get all TX options contracts
        contracts = []
        for option in self.api.Contracts.Options:
            for contract in option:
                if "TX" in contract.category:
                    contracts.append(
                        [
                            contract.symbol,
                            contract.name,
                            contract.update_date,
                            contract.delivery_date,
                        ]
                    )
        # Convert contracts data to DataFrame
        df = pd.DataFrame(
            contracts, columns=["symbol", "name", "update_date", "delivery_date"]
        )

        # Convert date columns to datetime format
        df["update_date"] = pd.to_datetime(df["update_date"])
        df["delivery_date"] = pd.to_datetime(df["delivery_date"])

        # Calculate date differences
        df["date_diff"] = (df["delivery_date"] - df["update_date"]).dt.days

        # Sort by date differences in ascending order
        df = df.sort_values(by="date_diff")
        return df

    def _get_next_contract(self, contracts_df, wednesday_time):
        now = datetime.datetime.now()

        if now < wednesday_time:
            return contracts_df.iloc[0]
        else:
            # Remove contracts with zero date difference
            contracts_df = contracts_df[contracts_df["date_diff"] != 0]
            contracts_df.reset_index(drop=True, inplace=True)
            return contracts_df.iloc[0]


class ShioajiWrapper:
    def __init__(self, api):
        self.api = api
        self.date_format = "%Y/%m/%d"
        self.target_time = (13, 45)

    def login(self, api_key, api_secret, contracts_timeout=10000):
        self.api.login(api_key, api_secret, contracts_timeout=contracts_timeout)

    def get_api(self):
        return self.api

    def get_stock_list(self):
        stock_list = [
            s
            for stock in self.api.Contracts.Stocks
            for s in stock
            if s["symbol"][:3] != "OES" and len(s["code"]) == 4
        ]
        code_list = sorted([i.code for i in stock_list])
        return code_list

    def get_snapshots(self, stock_list):
        contracts_dict = self.api.Contracts.Stocks
        contracts = [contracts_dict[i] for i in stock_list]
        snapshots = self.api.snapshots(contracts)
        df = pd.DataFrame(s.__dict__ for s in snapshots)
        df.ts = pd.to_datetime(df.ts)
        return df

    def is_after_target_time(self, dt):
        target_hour, target_minute = self.target_time
        return dt.time() >= datetime.time(target_hour, target_minute)

    def get_near_month_txf_contract(self):
        contract_list = [
            x for x in self.api.Contracts.Futures.TXF if x.code[-2:] not in ["R1", "R2"]
        ]

        now = datetime.datetime.now()
        contract_list = sorted(
            contract_list,
            key=lambda x: datetime.datetime.strptime(x.delivery_date, self.date_format),
        )

        first_contract_date = datetime.datetime.strptime(
            contract_list[0].delivery_date, self.date_format
        )

        if first_contract_date.date() == now.date() and self.is_after_target_time(now):
            return contract_list[1]

        return contract_list[0]

    # def get_near_month_txf_contract(self):
    #    contract = min(
    #        [
    #            x
    #            for x in self.api.Contracts.Futures.TXF
    #            if x.code[-2:] not in ["R1", "R2"]
    #        ],
    #        key=lambda x: x.delivery_date,
    #    )
    #    return contract

    def is_trading_day(self, now=datetime.datetime.now()):
        kbars = self.api.kbars(
            contract=self.api.Contracts.Indexs.TSE.TSE001,
            start=now.strftime("%Y-%m-%d"),
            end=now.strftime("%Y-%m-%d"),
        )
        return len(kbars.Open) > 0

    def get_latest_option_contract(self):
        contract_finder = ContractFinder(self.api)
        return contract_finder.get_latest_option_contract()
